import sys
from pathlib import Path

import cirq
import pytest
import qiskit.circuit.library as qcl
from qiskit import QuantumCircuit
from qiskit_finance.circuit.library.probability_distributions import (
    LogNormalDistribution,
    NormalDistribution,
)

import bluequbit

sys.path.append(Path(__file__).parent)
from bluequbit_test_lib import cirq_textbook_algorithms as algos  # noqa: E402

nqubits = 3

# Should be equivalent to the circuits being tested in deqart_backend.
circuits = [
    qcl.QFT(nqubits),
    qcl.QuantumVolume(nqubits),
    qcl.ExactReciprocal(2, 0.5),
    # Disabled for now because Qiskit has serialization roundtrip problem.
    # qcl.FourierChecking([1, -1, -1, -1], [1, 1, -1, -1]),
    qcl.IQP([[6, 5, 3], [5, 4, 5], [3, 5, 1]]),
]

# Grover circuit.
oracle = QuantumCircuit(2)
oracle.z(0)  # good state = first qubit is |1>
grover = qcl.GroverOperator(oracle, insert_barriers=True)
circuits.append(grover)

# Adder circuit
qc_adder = QuantumCircuit(3)
qc_adder.rz(0.5, 0)
qc_adder.rz(0.6, 1)
# 1 + 1 qubits
adder = qcl.DraperQFTAdder(1, kind="half")
qc_adder.append(adder, range(adder.num_qubits))
circuits.append(qc_adder)

# Multiplier circuit
qc_multiplier = QuantumCircuit(5)
qc_multiplier.rz(0.5, 0)
qc_multiplier.rz(0.6, 1)
# 1 x 1 qubits
mul = qcl.HRSCumulativeMultiplier(1)
qc_multiplier.append(mul, range(mul.num_qubits))
circuits.append(qc_multiplier)

circuits.extend(
    [
        NormalDistribution(3, mu=1, sigma=1, bounds=(0, 2)),
        LogNormalDistribution(
            [2, 2, 2], [1, 0.9, 0.2], [[1, -0.2, 0.2], [-0.2, 1, 0.4], [0.2, 0.4, 1]]
        ),
    ]
)

# Now we add Cirq circuits
circuits.extend(
    [algos.make_qft(3), algos.phase_estimation(0.123456, 3), algos.make_grover(3)]
)


@pytest.mark.parametrize("circuit", circuits)
def test_bq_run_circuit(circuit):
    bq = bluequbit.init()
    bq.run(circuit)


def test_temporary_validate_productofsums_cirq():
    qs = cirq.LineQubit.range(2)
    c = cirq.Circuit([cirq.ry(0.5).on(qs[1]).controlled_by(qs[0])])
    bq = bluequbit.init()
    bq.run(c)


def test_qiskit_cy():
    # See https://trello.com/c/0SMvJuLV/266-productofsums-issue
    qc_qiskit = QuantumCircuit(2)
    qc_qiskit.cy(0, 1)
    bq = bluequbit.init()
    bq.run(qc_qiskit)


def test_cirq_controlled_ry():
    # See https://trello.com/c/0SMvJuLV/266-productofsums-issue
    # Basically same as test_cirq_cy, except in Cirq.
    qs = cirq.LineQubit.range(2)
    circuit = cirq.Circuit([cirq.ry(0.5).on(qs[1]).controlled_by(qs[0])])
    bq = bluequbit.init()
    bq.run(circuit)


@pytest.mark.parametrize("seed", range(4))
def test_cirq_random_circuit(seed):
    qubits = [cirq.NamedQubit(f"q_{i}") for i in range(7)]
    circuit = cirq.testing.random_circuit(qubits, 5, 0.7, random_state=seed)
    bq = bluequbit.init()
    bq.run(circuit)


def test_cirq_circuit_with_non_terminal_measurements():
    qubits = cirq.LineQubit.range(2)
    circuit = cirq.Circuit()
    circuit.append(cirq.measure(qubits, key="measure1"))
    circuit.append(cirq.X(qubits[0]))
    circuit.append(cirq.X(qubits[1]))
    circuit.append(cirq.measure(qubits, key="measure2"))
    bq = bluequbit.init()
    with pytest.raises(bluequbit.exceptions.BQJobNotCompleteError) as e_info:
        bq.run(circuit)
        assert "FAILED_VALIDATION. Mid-circuit measurements" in str(e_info)


def test_cirq_circuit_with_terminal_measurements_only():
    qubits = cirq.LineQubit.range(2)
    circuit = cirq.Circuit()
    circuit.append(cirq.X(qubits[0]))
    circuit.append(cirq.X(qubits[1]))
    circuit.append(cirq.measure(qubits, key="measure1"))
    bq = bluequbit.init()
    bq.run(circuit)


@pytest.mark.parametrize("circuit", [None, 42, "42"])
def test_invalid_object(circuit):
    bq = bluequbit.init()
    with pytest.raises(Exception) as e_info:
        bq.run(circuit)
        assert "Unsupported circuit type" in str(e_info)


def test_empty_circuit():
    bq = bluequbit.init()
    circuit = cirq.Circuit()
    with pytest.raises(bluequbit.exceptions.BQJobNotCompleteError) as e_info:
        bq.run(circuit)
        assert "FAILED_VALIDATION. The circuit is empty" in str(e_info)



import os

import cirq
import numpy as np
import pytest

import bluequbit


@pytest.mark.skipif(
    "BLUEQUBIT_COUNT_BUG" not in os.environ,
    reason="Requires env variable to be set",
)
def test_count_wrong():
    bq = bluequbit.init()
    n = 26
    qc_cirq = cirq.testing.random_circuit(
        qubits=n, n_moments=50, op_density=1.0, random_state=1
    )

    measure_qubits = 1
    for _ in range(measure_qubits):
        qc_cirq.append(cirq.measure(cirq.NamedQubit("0")))

    device = "cpu"
    r0 = bq.run(
        qc_cirq, device=device, job_name=f"{n}_random_{measure_qubits}_{device}"
    )
    counts = r0.get_counts()
    assert len(counts) == 2
    sum_probs = 0.0
    for _, v in counts.items():
        sum_probs += v
    assert np.isclose(sum_probs, 1.0)



import logging

import packaging.version
import requests

logger = logging.getLogger("bluequbit-python-sdk")


def check_version(version):
    local_version = packaging.version.parse(version)
    if local_version.is_prerelease:
        logger.warning(
            "Development version %s of BlueQubit Python SDK is being used", version
        )
    req = requests.get("https://pypi.python.org/pypi/bluequbit/json", timeout=2.0)
    if not req.ok:
        return

    # find max version on PyPI
    releases = req.json().get("releases", [])
    pip_version = packaging.version.parse("0")
    for release in releases:
        ver = packaging.version.parse(release)
        if not ver.is_prerelease or local_version.is_prerelease:
            pip_version = max(pip_version, ver)

    if pip_version.major > local_version.major:
        logger.warning(
            "There is a major upgrade of BlueQubit Python SDK available on PyPI. We"
            " recommend upgrading. Run 'pip install --upgrade bluequbit' to upgrade"
            " from your version %s to %s.",
            local_version,
            pip_version,
        )
    elif pip_version > local_version:
        logger.info(
            "There is a newer version of BlueQubit Python SDK available on PyPI."
            " Run 'pip install --upgrade bluequbit' to upgrade from your version %s"
            " to %s.",
            local_version,
            pip_version,
        )



from . import job_metadata_constants


class BQError(Exception):
    def __init__(self, message="Unknown Error Message"):
        super().__init__(message)
        self.message = message


class BQSDKUsageError(BQError):
    def __init__(self, message):
        super().__init__(message)


class BQUnauthorizedAccessError(BQError):
    def __init__(self):
        super().__init__(
            "Wrong API key. You can find your API key once you log in to"
            " https://app.bluequbit.io."
        )


class BQJobStatevectorNotAvailableError(BQError):
    def __init__(self, job_id, num_qubits):
        super().__init__(
            "Job run with shots > 0 or statevector is too large for"
            f" {num_qubits} qubits (job: {job_id}). Please use .get_counts()"
            " instead.",
        )
        self.job_id = job_id
        self.num_qubits = num_qubits


class BQJobDoesNotContainCountsError(BQError):
    def __init__(self, job_id):
        super().__init__(f"The job {job_id} result metadata doesn't contain counts.")
        self.job_id = job_id


class BQJobNotCompleteError(BQError):
    def __init__(self, job_id, run_status, error_message):
        super().__init__(
            f"Job {job_id} finished with status: {run_status}. {error_message}",
        )
        self.job_id = job_id
        self.run_status = run_status
        self.error_message = error_message


class BQJobHasNotFinishedYetError(BQError):
    def __init__(self, job_id, run_status):
        super().__init__(
            f"Job {job_id} has not finished yet. Current status: {run_status}.",
        )
        self.job_id = job_id
        self.run_status = run_status


class BQJobCouldNotCancelError(BQError):
    def __init__(self, job_id, run_status, error_message):
        super().__init__(
            f"Couldn't cancel job {job_id}. Finished status is {run_status}."
            f" {error_message}",
        )
        self.job_id = job_id
        self.run_status = run_status


class BQBatchJobsLimitExceededError(BQError):
    def __init__(self, num):
        super().__init__(
            "Batch operations job count limit exceeded. The maximum is"
            f" {job_metadata_constants.MAXIMUM_NUMBER_OF_BATCH_JOBS}, but {num} was"
            " used."
        )
        self.num = num


class BQAPIError(BQError):
    def __init__(self, http_status_code, error_message):
        super().__init__(
            f"{error_message}. HTTP response status code: {http_status_code}.",
        )
        self.http_status_code = http_status_code
        self.error_message = error_message


class BQJobInvalidDeviceTypeError(BQError):
    def __init__(self, device):
        super().__init__(
            f"Invalid device type {device}. Must be one of"
            f" {', '.join(job_metadata_constants.DEVICE_TYPES)}."
        )



import threading

import qiskit

import bluequbit

NUM_PER_THREAD = 2
NUM_THREADS = 2

qc_qiskit = qiskit.QuantumCircuit(2)
qc_qiskit.h(0)
qc_qiskit.x(1)

job_ids = [[] for _ in range(NUM_THREADS)]


def submit_task(dq_client, thread_no):
    for _ in range(NUM_PER_THREAD):
        job_ids[thread_no].append(dq_client.run(qc_qiskit, asynchronous=True).job_id)


def test_stress_submit_jobs():
    dq_client = bluequbit.BQClient()
    threads = []
    for i in range(NUM_THREADS):
        threads.append(
            threading.Thread(
                target=submit_task,
                args=(
                    dq_client,
                    i,
                ),
            )
        )
        threads[-1].start()

    for i in range(NUM_THREADS):
        threads[i].join()

    job_ids_flat = [item for sublist in job_ids for item in sublist]

    assert len(job_ids_flat) == NUM_THREADS * NUM_PER_THREAD

    dq_client = bluequbit.BQClient()
    for job_id in job_ids_flat:
        job_result = dq_client.wait(job_id)
        assert job_result.run_status == "COMPLETED"
        assert job_result.error_message is None



import numpy as np
import pytest
import qiskit

import bluequbit

num_qubits = 30
qc_qiskit1 = qiskit.QuantumCircuit(num_qubits)
qc_qiskit1.x(np.arange(num_qubits))


@pytest.mark.flaky(reruns=3, reruns_delay=2)
def test_long_running_cancel_job():
    dq_client = bluequbit.BQClient()

    r = dq_client.run(qc_qiskit1, asynchronous=True)
    for _ in range(1200):
        res = dq_client.get(r.job_id)
        if res.run_status == "RUNNING":
            break
    else:
        raise AssertionError()

    r = dq_client.cancel(r.job_id)
    assert r.run_status == "CANCELED"
    assert r.error_message == "Canceled by user"



import datetime
import os

import numpy as np
import pytest
import qiskit

import bluequbit

qc_qiskit1 = qiskit.QuantumCircuit(2)
qc_qiskit1.h(0)
qc_qiskit1.x(1)
qc_qiskit2 = qiskit.QuantumCircuit(2)
qc_qiskit2.x(0)
qc_qiskit2.h(1)


def test_batch_jobs():
    time_now = datetime.datetime.now(datetime.timezone.utc)
    dq_client = bluequbit.BQClient()
    job_results = dq_client.run([qc_qiskit1, qc_qiskit2], job_name="testing_batch_jobs")
    expected = {
        ("testing_batch_jobs_0", job_results[0].job_id),
        (("testing_batch_jobs_1", job_results[1].job_id)),
    }
    results = dq_client.search(
        run_status="COMPLETED",
        created_later_than=time_now - datetime.timedelta(seconds=10),
    )
    num_found = 0
    for jr in results:
        if (jr.job_name, jr.job_id) in expected:
            num_found += 1
            assert jr.batch_id is not None
            expected.remove((jr.job_name, jr.job_id))

    assert num_found == 2


def test_batch_jobs_exceed_max():
    dq_client = bluequbit.BQClient()
    with pytest.raises(bluequbit.exceptions.BQBatchJobsLimitExceededError) as e_info:
        dq_client.run([qc_qiskit1] * 501, job_name="testing_batch_jobs")
    assert e_info.value.num == 501


def test_batch_job_estimation():
    dq_client = bluequbit.BQClient()
    results = dq_client.estimate([qc_qiskit1, qc_qiskit2])
    for result in results:
        assert result.estimated_runtime == 100


def test_batch_jobs_wait_get():
    dq_client = bluequbit.BQClient()
    job_results = dq_client.run(
        [qc_qiskit1, qc_qiskit2], job_name="testing_batch_jobs_async", asynchronous=True
    )
    expected = {
        ("testing_batch_jobs_async_0", job_results[0].job_id),
        (("testing_batch_jobs_async_1", job_results[1].job_id)),
    }
    results = dq_client.wait(
        [
            job_results[0].job_id,
            job_results[1].job_id,
        ]
    )
    num_found = 0
    for jr in results:
        if (jr.job_name, jr.job_id) in expected:
            num_found += 1
            expected.remove((jr.job_name, jr.job_id))

    assert num_found == 2

    expected = {
        ("testing_batch_jobs_async_0", job_results[0].job_id),
        (("testing_batch_jobs_async_1", job_results[1].job_id)),
    }
    results = dq_client.get(
        [
            job_results[0].job_id,
            job_results[1].job_id,
        ]
    )
    num_found = 0
    for jr in results:
        if (jr.job_name, jr.job_id) in expected:
            num_found += 1
            expected.remove((jr.job_name, jr.job_id))

    assert num_found == 2


num_qubits = 29
qc_qiskit3 = qiskit.QuantumCircuit(num_qubits)
qc_qiskit3.x(np.arange(num_qubits))

qc_qiskit4 = qiskit.QuantumCircuit(num_qubits)
qc_qiskit4.h(np.arange(num_qubits))


@pytest.mark.skipif(
    "BLUEQUBIT_LOCAL_ENV_TESTS" not in os.environ,
    reason="Requires env variable to be set",
)
@pytest.mark.flaky(reruns=3, reruns_delay=2)
def test_long_running_batch_cancel_job1():
    dq_client = bluequbit.BQClient()
    results = dq_client.run([qc_qiskit3, qc_qiskit4], asynchronous=True)
    for _ in range(1200):
        res = dq_client.get([results[0].job_id, results[1].job_id])
        if res[0].run_status == "RUNNING" and res[1].run_status == "RUNNING":
            break
    else:
        raise AssertionError()

    r1 = dq_client.cancel(results[0].job_id)
    assert r1.run_status == "CANCELED"
    assert r1.error_message == "Canceled by user"

    with pytest.raises(bluequbit.exceptions.BQJobNotCompleteError):
        # This test should be run on local setup, in that case the second job
        # will be terminated due to memory issue, which is the expected behavior
        dq_client.wait(results[1].job_id)


@pytest.mark.skipif(
    "BLUEQUBIT_LOCAL_ENV_TESTS" not in os.environ,
    reason="Requires env variable to be set",
)
@pytest.mark.flaky(reruns=3, reruns_delay=2)
def test_long_running_batch_cancel_job2():
    dq_client = bluequbit.BQClient()
    results = dq_client.run([qc_qiskit3, qc_qiskit4], asynchronous=True)
    for _ in range(1200):
        res = dq_client.get([results[0].job_id, results[1].job_id])
        if res[0].run_status == "RUNNING" and res[1].run_status == "RUNNING":
            break
    else:
        raise AssertionError()

    r1 = dq_client.cancel([results[0].job_id, results[1].job_id])
    for r in r1:
        assert r.run_status == "CANCELED"
        assert r.error_message == "Canceled by user"


@pytest.mark.skipif(
    "BLUEQUBIT_LOCAL_ENV_TESTS" not in os.environ,
    reason="Requires env variable to be set",
)
def test_batch_jobs_many():
    dq_client = bluequbit.BQClient()
    job_results = dq_client.run(
        [qc_qiskit1, qc_qiskit2] * 10, job_name="testing_batch_jobs"
    )
    for jr in job_results:
        assert jr.run_status in ["COMPLETED", "JOBS_LIMIT_EXCEEDED"]

    assert len(job_results) == 20



import pytest
import qiskit

import bluequbit


def test_get_job():
    dq_client = bluequbit.init()
    qc_qiskit = qiskit.QuantumCircuit(2)
    qc_qiskit.h(0)
    qc_qiskit.x(1)
    result = dq_client.run(qc_qiskit)

    assert result.num_qubits == 2
    print(result)

    assert result.get_statevector().shape == (4,)
    assert len(result.get_counts()) == 2
    assert sum(result.get_counts().values()) == 1.0


def test_get_job_counts():
    dq_client = bluequbit.init()
    qc_qiskit = qiskit.QuantumCircuit(2)
    qc_qiskit.h(0)
    qc_qiskit.x(1)
    qc_qiskit.measure_all()
    result = dq_client.run(qc_qiskit, shots=6)

    assert result.num_qubits == 2
    assert sum(result.get_counts().values()) == 6

    with pytest.raises(bluequbit.exceptions.BQJobStatevectorNotAvailableError) as e:
        result.get_statevector()
    assert (
        e.value.message
        == "Job run with shots > 0 or statevector is too large for"
        f" {result.num_qubits} qubits (job: {result.job_id}). Please use .get_counts()"
        " instead."
    )



import json
import logging
import os
from pathlib import Path

from .exceptions import BQAPIError, BQError, BQUnauthorizedAccessError
from .http_utils import request_retriable
from .version import __version__

logger = logging.getLogger("bluequbit-python-sdk")


class BackendConnection:
    def __init__(self, api_token=None):
        config_dir = Path.home() / ".config" / "bluequbit"
        config_location = config_dir / "config.json"

        main_endpoint_from_config_file = None
        ssl_verify_from_config_file = None
        token_from_env_variable = os.environ.get("BLUEQUBIT_API_TOKEN")

        if config_location.is_file():
            with config_location.open(encoding="utf-8") as f:
                config = json.load(f)
            main_endpoint_from_config_file = config.get("main_endpoint")
            ssl_verify_from_config_file = config.get("ssl_verify")

        if api_token is None:
            if token_from_env_variable is None:
                raise BQError(
                    "Please specify an API token to init(<your-api-token>) or have API"
                    " token set in BLUEQUBIT_API_TOKEN environment variable. You can"
                    " find your API token once you log in to https://app.bluequbit.io",
                )
            api_token = token_from_env_variable

        self._api_config = {
            "token": api_token,
            "main_endpoint": (
                "https://app.bluequbit.io/api/v1"
                if main_endpoint_from_config_file is None
                else main_endpoint_from_config_file
            ),
            "ssl_verify": (
                True
                if ssl_verify_from_config_file is None
                else ssl_verify_from_config_file
            ),
        }

        try:
            self._token = self._api_config["token"]
        except KeyError:
            raise BQError(
                "Incorrect config file: 'token' key is not present in the config"
                f" file {config_location}.",
            ) from None

        self._default_headers = {
            "Authorization": f"SDK {self._token}",
            "Connection": "close",
            "User-Agent": f"BlueQubit SDK {__version__}",
        }

        self._main_endpoint = "https://app.bluequbit.io/api/v1"
        if (
            "main_endpoint" in self._api_config
            and self._main_endpoint != self._api_config["main_endpoint"]
        ):
            self._main_endpoint = self._api_config["main_endpoint"]
            logger.warning("Using custom endpoint %s", self._main_endpoint)
        self._verify = True
        if "ssl_verify" in self._api_config:
            self._verify = self._api_config["ssl_verify"]
        self._session = None
        self._authenticated = False
        response = self.send_request(
            req_type="GET",
            path="/jobs",
            params={"limit": 1},
            authentication_request=True,
        )

        if response.ok:
            self._authenticated = True
        elif response.status_code == 401:
            raise BQUnauthorizedAccessError
        else:
            raise BQAPIError(
                response.status_code, f"Couldn't reach BlueQubit {response.text}."
            )

    @staticmethod
    def update_config_file(content, config_location):
        with config_location.open("w", encoding="utf-8") as f:
            json.dump(content, f, indent=4)
        logger.info("Configuration file %s successfully updated.", config_location)

    def send_request(
        self,
        req_type,
        path,
        params=None,
        data=None,
        json_req=None,
        headers=None,
        authentication_request=False,
    ):
        if not authentication_request and not self._authenticated:
            raise BQError(
                "BQClient was not initialized. Please provide correct API token to"
                " BQClient(<token>).",
            )
        url = self._main_endpoint + path

        if params is not None:
            for key, value in params.items():
                if isinstance(value, str):
                    params[key] = value.replace("\\", "\\\\")
                if isinstance(value, list):
                    params[key] = ",".join(value)

        if headers is None:
            headers_to_send = self._default_headers
        else:
            headers_to_send = dict(self._default_headers, **headers)

        resp = request_retriable(
            method=req_type,
            url=url,
            data=data,
            json=json_req,
            params=params,
            headers=headers_to_send,
        )
        return resp
